############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

import types, weakref
from StringIO import StringIO
from ZSI.wstools.WSDLTools import GetDocumentation, ParseQName, WSDLReader, \
    PortType, WSDL, Message
from ZSI.wstools.XMLSchema import XMLBase, SplitQName
from ZSI.wstools.Utility import DOM, Collection, DOMException

class GWSDLReader(WSDLReader):
    """A WSDLReader creates GWSDL instances from urls and xml data."""

    # Custom subclasses of WSDLReader may wish to implement a caching
    # strategy or other optimizations. Because application needs vary 
    # so widely, we don't try to provide any caching by default.

    def loadFromStream(self, file):
        """Return a WSDL instance loaded from a file object."""
        document = DOM.loadDocument(file)
        wsdl = GWSDL()
        wsdl.location = file.name
        wsdl.load(document)
        return wsdl

    def loadFromURL(self, url):
        """Return a WSDL instance loaded from the given url."""
        document = DOM.loadFromURL(url)
        wsdl = GWSDL()
        wsdl.location = url
        wsdl.load(document)
        return wsdl

    def loadFromString(self, data):
        """Return a WSDL instance loaded from an xml string."""
        return self.loadFromStream(StringIO(data))

    def loadFromFile(self, filename):
        """Return a WSDL instance loaded from the given file."""
        file = open(filename, 'rb')
        try:
            wsdl = self.loadFromStream(file)
        finally:
            file.close()
        return wsdl

class GWSDL(WSDL):
    """A WSDL object models a WSDL service description. WSDL objects
       may be created manually or loaded from an xml representation
       using a WSDLReader instance."""

    def __init__(self, targetNamespace=None):
        WSDL.__init__(self, targetNamespace)
        self.gPortTypes = Collection(self)

    def addMessage(self, name, documentation=''):
        #XXX THIS IS NOT A SOLUTION
        #if self.messages.has_key(name):
        #    raise WSDLError(
        #        'Duplicate message element: %s.' % name
        #        )
        item = Message(name, documentation)
        self.messages[name] = item
        return item

    def load(self, document):
        # We save a reference to the DOM document to ensure that elements
        # saved as "extensions" will continue to have a meaningful context
        # for things like namespace references. The lifetime of the DOM
        # document is bound to the lifetime of the WSDL instance.
        WSDL.load(self, document)

        for indx in range(len(self.extensions)):
            element = self.extensions[indx]
            localName = element.localName
            namespace = element.namespaceURI

            if namespace == GWSDLComponent.namespace:
                if localName == 'portType':
                    name = DOM.getAttr(element, 'name')
                    docs = GetDocumentation(element)
                    self.gPortTypes[name] = GPortType(self, name, docs)
                    self.gPortTypes[name].load(element)
                    self.extensions[indx] = None
                else:
                    raise GWSDLError, 'unknown component %s' % name


class GWSDLError(Exception): pass

class Any:
    def __init__(self, namespace, minOccurs, maxOccurs):
        self.namespace = namespace
        self.minOccurs = minOccurs
        self.maxOccurs = maxOccurs

    
class AnyAttribute:
    def __init__(self, namespace, minOccurs, maxOccurs):
        self.namespace = namespace
        self.minOccurs = minOccurs
        self.maxOccurs = maxOccurs
    

class GWSDLComponent(XMLBase):
    namespace = 'http://www.gridforum.org/namespaces/2003/03/gridWSDLExtensions'
    required = []
    attributes = {}
    contents = {}
    xmlns_key = ''
    xmlns = 'xmlns'
    xml = 'xml'
    any = 'any'
    anyAttribute = 'anyAttribute'

    def __init__(self, parent=None):
        """parent -- parent instance
           instance variables:
               attributes -- dictionary of node's attributes
        """
        self.attributes = None
        self.parent = parent
        if self.parent:
            self.parent = weakref.ref(parent)

        if not self.__class__ == GWSDLComponent\
           and not (type(self.__class__.required) == type(GWSDLComponent.required)\
           and type(self.__class__.attributes) == type(GWSDLComponent.attributes)\
           and type(self.__class__.contents) == type(GWSDLComponent.contents)):
            raise RuntimeError, 'Bad type for a class variable in %s' %self.__class__

    def getAttribute(self, attribute):
        """return requested attribute or None
        """
        return self.attributes.get(attribute)

    def setAttributes(self, node):
        """Sets up attribute dictionary, checks for required attributes and 
           sets default attribute values. attr is for default attribute values 
           determined at runtime.
           
           structure of attributes dictionary
               ['xmlns'][xmlns_key] --  xmlns namespace
               ['xmlns'][prefix] --  declared namespace prefix 
               [namespace][prefix] -- attributes declared in a namespace
               [attribute] -- attributes w/o prefix, default namespaces do
                   not directly apply to attributes, ie Name can't collide 
                   with QName.
        """
        self.attributes = {GWSDLComponent.xmlns:{}}

        #for key in self.__class__.attributes.keys():
        for v in node._attrs.values():
            key = v.nodeName
            value = v.nodeValue
            
            if key == GWSDLComponent.xmlns:
                self.attributes[GWSDLComponent.xmlns][GWSDLComponent.xmlns_key] = value
            elif SplitQName(key)[1] == GWSDLComponent.xmlns:
                self.attributes[GWSDLComponent.xmlns][SplitQName(key)[0]] = value
            elif value: 
                try:
                    ns, kvalue = ParseQName(key, node)
                except DOMException, ex:
                    raise
                if ns:
                    if not self.attributes.has_key(ns):
                        self.attributes[ns] = {}
                    elif self.attributes[ns].has_key(kvalue):
                         raise GWSDLError(
                             'attribute %s,%s declared multiple times' %(ns,kvalue)
                             )
                    self.attributes[ns][kvalue] = value
                if not self.attributes.has_key(key):
                    self.attributes[key] = value
                else:
                    raise GWSDLError(
                        'attribute %s declared multiple times' %value
                        )

        #self.__checkAttributes()
        self.__setAttributeDefaults()

        key = 'extends'
        if self.attributes.has_key(key):
            extends = self.attributes[key].split()
            self.attributes[key] = []
            for pt in extends:
                nsuri, name = ParseQName(pt, node)
                self.attributes[key].append((nsuri, name))

    def __setAttributeDefaults(self):
        """Looks for default values for unset attributes.  If
           class variable representing attribute is None, then
           it must be defined as an instance variable.
        """
        for k,v in self.__class__.attributes.items():
            if v and not self.attributes.has_key(k):
                if isinstance(v, types.FunctionType):
                    self.attributes[k] = v(self)
                else:
                    self.attributes[k] = v

    def __checkAttributes(self):
        """Checks that required attributes have been defined,
           attributes w/default cannot be required.   Checks
           all defined attributes are legal, attribute 
           references are not subject to this test.
        """
        for a in self.__class__.required:
            if not self.attributes.has_key(a):
                raise GWSDLError,\
                    'class instance %s, missing required attribute %s'\
                    %(self.__class__, a)

        for a in self.attributes.keys():
            if (a != GWSDLComponent.xmlns) and\
                (a not in self.__class__.attributes.keys()):
                #and not (self.isAttribute() and self.isReference()):
                any = self.__class__.attributes.get(GWSDLComponent.anyAttribute)
                if not any:
                    raise GWSDLError, '%s, unknown attribute' %a



class GPortType(GWSDLComponent, PortType):
    """<portType>
       parents:
           definitions
       attributes:
           name -- NCName, required
           extends -- list of QNames
       contents:
           operation
           serviceData
           staticServiceDataValues
    """
    required = ['name']
    attributes = {'name':None, 'type':None, 'minOccurs':1, 'maxOccurs':1,
        'mutability':True, 'modifiable':True, 'nillable':True , 'extends':None}
    contents = {}

    def __init__(self, parent, name, documentation=''):
        PortType.__init__(self, name=name, documentation=documentation)
        GWSDLComponent.__init__(self, parent)
        self.serviceData = Collection(self)
        self.staticServiceDataValues = Collection(self)

    def getPortTypeExtensions(self):
        wsdl = self.getWSDL()
        extends = []
        for gpt in self.attributes.get('extends', []):
            nsuri, name = gpt
            #if nsuri == wsdl.targetNamespace:
            #    extends.append(wsdl.portTypes[name])
            if wsdl.portTypes.has_key(name):
                extends.append(wsdl.portTypes[name])
            elif wsdl.gPortTypes.has_key(name):
                extends.append(wsdl.gPortTypes[name])
            else:
                raise GWSDLError, 'Parent wsdl has no %s PortType' %name
        return extends

    def load(self, element):
        self.setAttributes(element)
        #extends = DOM.getAttr(element, 'extends')
        #for pt in extends.split():
        #    nsuri, name = ParseQName(pt, element)
        #    self.extends.append((nsuri,name))

        #self.extends = DOM.getAttr(element, 'name')        
        PortType.load(self, DOM.getElements(element, 'operation'))
        for element in DOM.getElements(element, 'serviceData'):
            component = ServiceDataDeclaration()
            component.load(element)

            #name = DOM.getAttr(element, 'name')
            #docs = GetDocumentation(element)
            #param_order = DOM.getAttr(element, 'parameterOrder', default=None)
            #if param_order is not None:
            #    param_order = param_order.split(' ')
            #serviceData = self.addServiceData(name, docs, param_order)
            name =  component.getAttribute('name')
            self.serviceData[name] = component

            item = DOM.getElement(element, 'input', None, None)
            if item is not None:
                name = DOM.getAttr(item, 'name')
                docs = GetDocumentation(item)
                msgref = DOM.getAttr(item, 'message')
                message = msgref.split(':', 1)[-1]
                operation.setInput(message, name, docs)

            item = DOM.getElement(element, 'output', None, None)
            if item is not None:
                name = DOM.getAttr(item, 'name')
                docs = GetDocumentation(item)
                msgref = DOM.getAttr(item, 'message')
                message = msgref.split(':', 1)[-1]
                operation.setOutput(message, name, docs)

            for item in DOM.getElements(element, 'fault', None):
                name = DOM.getAttr(item, 'name')
                docs = GetDocumentation(item)
                msgref = DOM.getAttr(item, 'message')
                message = msgref.split(':', 1)[-1]
                operation.addFault(message, name, docs)


class ServiceDataDeclaration(GWSDLComponent):
    """<serviceData>
       parent:
           portType
       attributes:
           name -- NCName, Required
           type -- QName
           minOccurs -- Whole Number, 1
           maxOccurs -- (Whole Number | 'unbounded'), 1
           mutability -- ('static', 'constant', 'extendable', 'mutable'), 
                          'extendable'
           modifiable -- boolean, 'false'
           nillable -- boolean, 'false'
           anyAttribute -- ##other (any attributes w/non-schema namespace)
       contents:
           any -- ##any, (0, 'unbounded')
    """
    required = ['name']
    attributes = {
        'name':None, 'type':None, 'minOccurs':1, 'maxOccurs':1,
        'mutability':True, 'modifiable':True, 'nillable':True, 
        GWSDLComponent.anyAttribute:AnyAttribute('##any',0,'unbounded')
        }
    contents = {
        GWSDLComponent.any:Any('##any', 0, 'unbounded')
        }

    def __init__(self):
        self.attributes = None
        self.contents = None

    def load(self, element):
        self.attributes = {}
        self.contents = []

        #docs = GetDocumentation(element)
        #param_order = DOM.getAttr(element, 'parameterOrder', default=None)
        #if param_order is not None:
        #    param_order = param_order.split(' ')

        self.setAttributes(element)
        for element in DOM.getElements(element, None):
            localName = element.localName
            namespace = element.namespaceURI

            for k,v in self.__class__.contents.items():
                self.contents.append()

class ServiceDataValuesType(GWSDLComponent):
    """ServiceDataValuesType
       parent:
           gwsdl:portType
       attributes:
       contents:
           any -- ##any, (0, 'unbounded')
    """
    contents = {GWSDLComponent.any:Any('##any', 0, 'unbounded')}

    def __init__(self):
        self.attributes = None
        self.contents = None
 
    def load(self, element):
        self.attributes = {}
        self.contents = []

        for element in DOM.getElements(element, None):
            #localName = element.localName
            #namespace = element.namespaceURI
            any = self.__class__.contents[GWSDLComponent.any]
            #component = any.load(element)
            self.contents.append(element)


class ServiceDataValues(ServiceDataValuesType):
    pass

class StaticServiceDataValues(ServiceDataValuesType):
    pass
