############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import sys, types

from xml.dom.ext import c14n
from xml.dom.ext.reader import PyExpat
from pyGlobus.security import GSSCred, GSSContext, GSSContextException
from pyGlobus.security import GSSName, GSSMechs, GSSUsage, GSSBuffer
from pyGlobus.security import ContextRequests
from pyGridWare.ogsi.services.SecureContextEstablishment_services import SecureContextEstablishmentServiceLocator
#from Authenticator import Authenticator

from pyGridWare import getLogger
logger = getLogger(__name__)


class ExtendedGSSContext:
    """
    """
    def __init__(self, cred):
        """cred -- GSSCred
        """
        self._cred = cred
        lifetime, credName = self.inquireCred()
        self._name = credName
        self._context = GSSContext()
        self._requests = ContextRequests()

    def inquireCred(self):
        lifetime, credName = self._cred.inquire_cred()
        return lifetime, credName

    def setDelegation(self):
        self._requests.set_delegation()

    def hashCode(self):
        """init context-id
        """
        return '00000000-01a9-cf52-00000000-000000000000'

    def initSecContext(self, inToken=None):
        """XXX pyGlobus throws a segmentation Fault if inToken=None

           The ContextRequests object needs to be the last argument to the
           init_context calls that you make.
        """
        if inToken:
            major,minor,outToken = \
                self._context.init_context(\
                    init_cred=self._cred, 
                    target_name=self._name, 
                    inputTokenString=inToken, 
                    requests=self._requests)
        else:
            major,minor,outToken = \
                self._context.init_context(\
                    init_cred=self._cred, 
                    target_name=self._name,
                    inputTokenString="",
                    requests=self._requests)
        return outToken

    def getMIC(self, signedInfoC14n):
        signatureValue = self._context.get_mic(signedInfoC14n)
        return signatureValue

    def verifyMIC(self, signedInfoC14n, signatureValue):
        return self._context.verify_mic(signedInfoC14n, signatureValue)


class SecurityManager: pass


class GSSManager(SecurityManager):
    """
    """
    def createCredential(self):
        cred = GSSCred()
        name, mechs, usage  = GSSName(), GSSMechs(), GSSUsage()
        usage.set_usage_initiate()
        cred.acquire_cred(name, mechs, usage)
        return cred

    def createContext(self, cred):
        context = ExtendedGSSContext(cred)
        return context
           

class GSIConstants:
    GSI_CREDENTIALS = 1
    DELEGATION_TYPE = 2
    DELEGATION_TYPE_LIMITED = 3


class SOAPMessage:
    def __init__(self, message, context, URL):
        self._reader = PyExpat.Reader()
        self._doc = None
        self._secCtx = context
        self._cred = None
        self._url = URL
        self.setMessage(message)

    def getURL(self):
        return self._url

    def getContext(self):
        #return self._secCtx.getContext()
        return self._secCtx

    def getContextID(self):
        return self._secCtx.getContextID()

    def getMessageDocument(self):
        return self._doc

    def getMessageString(self):
        return c14n.Canonicalize(self._doc)

    def setCredential(self, cred):
        self._cred = cred

    def setContext(self, context):
        self._secCtx = context

    def setDelegation(self):
        self._secCtx.setDelegation()
 
    def setMessage(self, msg):
        if type(msg) == types.StringType:
            self._doc = self._reader.fromString(msg)
        else:
            self._doc = msg


class SecContext:
    """Associates a GSSContext with a context Id. It is used for GSI secure converstation.
    """

    def __init__(self):
        self._contextID = None
        self._cred = None
        self._context = None

        manager = GSSManager()
        cred = manager.createCredential()
        context = manager.createContext(cred)
        self.setContext(context)
        self.setCredential(cred)

    def getContext(self):
        """org.ietf.jgss.GSSContext getContext()
        """
        return self._context

    def getContextID(self):
        """java.lang.String getContextId()
        """
        return self._contextID

    def getDelegatedCredential(self):
        """org.ietf.jgss.GSSCredential getDelegatedCredential() 
        """
        self._cred

    def setCredential(self, cred):
        self._cred = cred

    def setContext(self, context):
        self._context = context

    def setContextID(self, contextID):
        self._contextID = contextID

    def setDelegation(self):
        self._context.setDelegation()


