############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import sha, base64
from xml.ns import SOAP, DSIG, XMLNS
from Ft.Xml import XPath
from pyGridWare.ogsi.security.SignedInfo import SignedInfo
from pyGridWare.ogsi.security import WSSE, WSU

class GssSignedInfo(SignedInfo):
    SignatureMethod = "http://www.globus.org/2002/04/xmlenc#gssapi-sign"
    refExp = XPath.Compile('soapenv:Envelope/soapenv:Header/wsse:Security/ds:Signature/ds:SignedInfo/ds:Reference')
    sigExp = XPath.Compile('soapenv:Envelope/soapenv:Header/wsse:Security/ds:Signature')
    secExp = XPath.Compile('soapenv:Envelope/soapenv:Header/wsse:Security')
    headExp = XPath.Compile('soapenv:Envelope/soapenv:Header')
    digExp = XPath.Compile('//*[@wsu:Id="digestSource"]')

    def __init__(self, doc):
        SignedInfo.__init__(self, doc)
        xContext = XPath.Context.Context(self._doc, processorNss=self.processorNss)
        if len(self.siExp.evaluate(xContext)) == 0:
            self.setUp()

    def setUp(self):
        """XXX not automatically prefixed??  
           SOAP-ENC prefix
        """
        xContext = XPath.Context.Context(self._doc, processorNss=self.processorNss)
        for node in self.headExp.evaluate(xContext):
            break
        else:
            envelopeNode = self._doc.getElementsByTagNameNS(SOAP.ENV, 'Envelope')[0]
            refChild = self._doc.getElementsByTagNameNS(SOAP.ENV, 'Body')[0]
            node = self._doc.createElementNS(SOAP.ENV, "SOAP-ENV:Header")
            #envelopeNode.appendChild(node)
            envelopeNode.insertBefore(newChild=node, refChild=refChild)
        headerNode = node
        
        securityNode = self._doc.createElementNS(WSSE.BASE, "wsse:Security")
        securityNode.setAttributeNS(XMLNS.BASE, 'xmlns:wsse', WSSE.BASE)
        securityNode.setAttributeNS(namespaceURI=SOAP.ENV, qualifiedName="SOAP-ENV:actor", value="")
        securityNode.setAttributeNS(namespaceURI=SOAP.ENV, qualifiedName="SOAP-ENV:mustUnderstand", value="0")
        headerNode.appendChild(securityNode)

        signatureNode = self._doc.createElementNS(DSIG.BASE, "ds:Signature")
        securityNode.setAttributeNS(XMLNS.BASE, 'xmlns:ds', DSIG.BASE)
        securityNode.appendChild(signatureNode)

        siNode = self._doc.createElementNS(DSIG.BASE, "ds:SignedInfo")
        signatureNode.appendChild(siNode)

        keyInfoNode = self._doc.createElementNS(DSIG.BASE, "ds:KeyInfo")
        signatureNode.appendChild(keyInfoNode)

        canMethodNode = self._doc.createElementNS(DSIG.BASE, "ds:CanonicalizationMethod")
        canMethodNode.setAttribute(name="Algorithm", value=self.getCanonicalizationMethodURI())
        sigMethodNode = self._doc.createElementNS(DSIG.BASE, "ds:SignatureMethod")
        sigMethodNode.setAttribute(name="Algorithm", value=self.getSignatureMethodURI())
        for node in (canMethodNode, sigMethodNode,):
            siNode.appendChild(node)

    def generateDigestValues(self):
        """XXX HARDCODED #digestSource. This is only set up to handle one digestSource.
        """
        xContext = XPath.Context.Context(self._doc, processorNss=self.processorNss)
        siNode = self.siExp.evaluate(xContext)[0]
        for node in self.digExp.evaluate(xContext):
            refNode = self._doc.createElementNS(DSIG.BASE, "ds:Reference")
            refNode.setAttribute(name="URI", value="#digestSource")
            siNode.appendChild(refNode)

            digestMethodNode = self._doc.createElementNS(DSIG.BASE, "ds:DigestMethod")
            digestMethodNode.setAttribute(name="Algorithm", value="http://www.w3.org/2000/09/xmldsig#sha1")
            refNode.appendChild(digestMethodNode)

            cxml = self._canonicalize(node)
            digest = sha.sha(cxml).digest()
            digestValue = base64.encodestring(digest).strip()

            digestNode = self._doc.createElementNS(DSIG.BASE, "ds:DigestValue")
            digestNode.appendChild(self._doc.createTextNode(digestValue))
            refNode.appendChild(digestNode)

    def verify(self):
        verify = False
        xContext = XPath.Context.Context(self._doc, processorNss=self.processorNss)

        #Find <XXX wsu:Id="digestSource">
        digExp = XPath.Compile('//*[@wsu:Id]')
        for signedNode in digExp.evaluate(xContext):
           attrID = signedNode.getAttributeNodeNS(self.processorNss['wsu'],'Id')
           key = attrID.value

           #Find <ds:Reference URI="#digestSource">
           xpath = 'soapenv:Envelope/soapenv:Header/wsse:Security/ds:Signature/ds:SignedInfo/ds:Reference[@URI="#%s"]' %key
           exp = XPath.Compile(xpath)
           nodes = exp.evaluate(xContext)
           if len(nodes) != 1:
               raise RuntimeError, 'SignatureInfo Reference URI %s must be unique' %key

           #Get DigestMethod and DigestValue
           xpathM = xpath + '/ds:DigestMethod/@Algorithm' 
           exp = XPath.Compile(xpathM)
           attrNode = exp.evaluate(xContext)[0]
           digestM = attrNode.value

           xpathV = xpath + '/ds:DigestValue/text()'
           exp = XPath.Compile(xpathV)
           chNode = exp.evaluate(xContext)[0]
           digestV = chNode.data

           base64.decodestring(digestV)
           cxml = self._canonicalize(signedNode)
           digest = sha.sha(cxml).digest()
           if digest != base64.decodestring(digestV):
               break
        else:
            verify = True
        return verify

    def verifySignatureValue(self, context):
        """context -- GSSContext
        """
        xContext = XPath.Context.Context(self._doc, processorNss=self.processorNss)

        #Get Canonicalized SignedInfo
        siNode = self.siExp.evaluate(xContext)[0]
        siC14n = self._canonicalize(siNode)

        #Get SignatureValue
        exp = XPath.Compile('soapenv:Envelope/soapenv:Header/wsse:Security/ds:Signature/ds:SignatureValue/text()')
        sigValue = base64.decodestring(exp.evaluate(xContext)[0].data)

        return context.verifyMIC(siC14n, sigValue)


