#!/usr/bin/env python

############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

import sys, os, unittest
import socket, urlparse
from ConfigParser import ConfigParser
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.wstools.TimeoutSocket import TimeoutError
from ZSI.wsdl2python import WriteServiceModule
from pyGridWare.ogsi.security.SecurityHandler import AuthenticationHandler, \
    SignContentsHandler, SignatureHandler

CONFIG_FILE = 'config.txt'
SEC_LOCATION = 'Location'
SEC_PATHS = 'Paths'
SEC_REFERENCES = 'schemaPath'
TEST = 'NotificationFactoryService'
MODULE_DIR = 'stubs'

try:
    os.mkdir(MODULE_DIR)
except OSError, ex:
    pass
sys.path.append(MODULE_DIR)

"""This test file generates the types and service files, then runs
   several tests.
"""

class Holder: 
    """This is just for specifying an element w/o content in
       an instance.
    """
    pass


class ServiceTestCase(unittest.TestCase):
    """Must set these class variables typeModuleName, serviceModuleName,
       path, and schema.  moduleDict is filled in after modules are created.
    """
    netloc = None
    typeModuleName = None
    serviceModuleName = None
    path = None
    schemaPath = None
    port = None
    moduleDict = {}

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        if not (self.path and self.schemaPath):
            raise RuntimeError, 'class variables must be initialized'

    def setUp(self):
        """Generate types and services modules if no record of them
           in the moduleDict class variable.
        """
        if not self.moduleDict.has_key(self.typeModuleName):
            os.chdir(MODULE_DIR)
            reader = WSDLReader()
            wsdl = reader.loadFromURL(self.schemaPath)
            wsm = WriteServiceModule(wsdl, 'pyGridWare.ogsi.Decorator', 'OGSIDecorator')
            wsm.write(False)
            self.setModuleNames(*wsm.get_module_names())
            os.chdir('../')
        for name in (self.typeModuleName, self.serviceModuleName):
            exec('import %s' %name)
            self.moduleDict[name] = eval(name)

    def getRequest(self):
        return getattr(self.moduleDict[self.serviceModuleName], self.requestName)()

    def setSecurity(self, port):
        auth = AuthenticationHandler()
        auth.setDelegation()
        port.binding.addSecurityHandler(auth)
        port.binding.addSecurityHandler(SignContentsHandler())
        port.binding.addSecurityHandler(SignatureHandler())

    def setModuleNames(self, typesModuleName, serviceModuleName):
        ServiceTestCase.typeModuleName = typesModuleName
        ServiceTestCase.serviceModuleName = serviceModuleName

    def test(self):
        raise NotImplementedError, 'not implemented'


class FactoryServiceTestCase(ServiceTestCase):
    #serviceModuleName = 'NotificationFactoryService_services'
    #typeModuleName = 'NotificationFactoryService_services_types'
    #path = '/ogsa/services/samples/counter/basic/CounterFactoryService'
    #schema = '/usr/local/java/gt3.2-core-src/impl/java/schema/ogsi/ogsi_notification_factory_service.wsdl'
    #port = None

    def __init__(self, methodName):
        ServiceTestCase.__init__(self, methodName)

    def setUp(self):
        ServiceTestCase.setUp(self)
        if not self.port:
            locator = self.moduleDict[self.serviceModuleName].NotificationFactoryServiceLocator()
            kw = {'tracefile':sys.stdout, 'url':self.path}
            url = urlparse.urlunparse(('http', self.netloc, self.path, '', '', ''))
            FactoryServiceTestCase.port = locator.getNotificationFactory(portAddress=url, **kw)
            self.setSecurity(self.port)

    def test(self):
        request = self.getRequest()
        self.setRequest(request)
        response = getattr(self.port, self.operationName)(request)
        self.checkResponse(response)

    def setRequest(self, request):
        raise NotImplementedError, 'not implemented'

    def checkResponse(self, msg):
        raise NotImplementedError, 'not implemented'


class CreateServiceTestCase(FactoryServiceTestCase):
    requestName = 'CreateServiceInputMessageWrapper'
    operationName = 'createService'

    def __init__(self, methodName):
        FactoryServiceTestCase.__init__(self, methodName)

    def setRequest(self, request):
        """Setting these variables to an empty instance so 
           that an empty element tag is produced.
           ie. <creationParameters/>
        """
        request._terminationTime = Holder()
        request._creationParameters = Holder()

    def checkResponse(self, msg):
        print "RESPONSE: ", msg
        print "\tlocator = ", msg._locator
        print "\treference = ", msg._locator._reference
        print "\tdefinitions = ", msg._locator._reference._definitions
        print "\tcurrentTerminationTime = ", msg._currentTerminationTime
        print "\textensibilityOutput = ", msg._extensibilityOutput


class FindServiceDataTestService(FactoryServiceTestCase):
    """
    """
    requestName = 'FindServiceDataInputMessageWrapper'
    operationName = 'findServiceData'

    def __init__(self, methodName):
        FactoryServiceTestCase.__init__(self, methodName)

    def setRequest(self, request):
        """request._queryExpression._name
               values: subscribeExtensibility, name, entry
        """
        request.setSerializationElement(variable='queryExpression') 
        request._queryExpression.setSerializationElement(variable='queryByServiceDataNames', 
            typecode=self.moduleDict[self.typeModuleName].ns1.queryByServiceDataNames_Dec)
        request._queryExpression._queryByServiceDataNames._name = 'entry'

    def checkResponse(self, msg):
        """serviceDataValues -- will default to a DOM element node because
               the serviceData namespace isn't available.  If want this
               available need to create the serviceData types module and
               import it in this script.
        """
        print "RESPONSE: ", msg
        print "\tresult -- ", msg._result
        print "\t\tserviceDataValues -- ", msg._result._serviceDataValues


def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FindServiceDataTestService))
    suite.addTest(unittest.makeSuite(CreateServiceTestCase))
    return suite


def main():
    cp = ConfigParser()
    cp.read(CONFIG_FILE)
    location = cp.get(SEC_LOCATION, TEST)
    if not location:
        host = socket.gethostbyaddr(socket.gethostname())[2][0]
        port = 8080
        ServiceTestCase.netloc = '%s:%d'%(host,port)
    else:
        host,port = eval(location)
        ServiceTestCase.netloc = '%s:%d'%(host,port)
    path = cp.get(SEC_PATHS, TEST)
    if path:
        ServiceTestCase.path = path
    schemaPath = cp.get(SEC_REFERENCES, TEST)
    if schemaPath:
        urlTuple = ('http',ServiceTestCase.netloc,'/%s' %schemaPath,'','','')
        ServiceTestCase.schemaPath = urlparse.urlunparse(urlTuple)
    testProgram = unittest.TestProgram(defaultTest="makeTestSuite")


if __name__ == '__main__':
    main()
